from __future__ import absolute_import
import datetime as dt
from decimal import Decimal
import six
from six.moves import filter, map, range
from sqlalchemy import event
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.event import listens_for
from sqlalchemy.schema import DDL, DropIndex
from flask import Markup, current_app, url_for
from flask_babel import gettext, lazy_gettext
from flask_login import current_user

from . import db
from .util import DeclEnum, classproperty, AutoID, Timestamped, AutoName,\
        unistr, ensure_unicode, PrettyDecimal, PrettyNumeric, DateTime
from .auth import PermissionType


if six.PY3:
    unicode = str

class ActionType(DeclEnum):

    # The actual stored values are single character to make it easier on
    # engines that don't support native enum types.

    # TRANS: Name of the status a request is in when it has been submitted and
    # TRANS: is ready to be evaluated.
    evaluating = u'evaluating', lazy_gettext(u'Evaluating')
    """Status for a request being evaluated."""

    # TRANS: Name of the status where a request has had a payout amount set,
    # TRANS: and is ready to be paid out. In other words, approved for payout.
    approved = u'approved', lazy_gettext(u'Approved')
    """Status for a request that has been evaluated and is awaitng payment."""

    # TRANS: Name of the status a request is in if the ISK has been sent to the
    # TRANS: requesting person, and no further action is needed.
    paid = u'paid', lazy_gettext(u'Paid')
    """Status for a request that has been paid. This is a terminatint state."""

    # TRANS: Name of the status a request has where a reviewer has rejected the
    # TRANS: request for SRP.
    rejected = u'rejected', lazy_gettext(u'Rejected')
    """Status for a requests that has been rejected. This is a terminating
    state.
    """

    # TRANS: When a request needs more information to be approved or rejected,
    # TRANS: it is in this status.
    incomplete = u'incomplete', lazy_gettext(u'Incomplete')
    """Status for a request that is missing details and needs further
    action.
    """

    # TRANS: A comment made on a request.
    comment = u'comment', lazy_gettext(u'Comment')
    """A special type of :py:class:`Action` representing a comment made on the
    request.
    """

    @classproperty
    def finalized(cls):
        return frozenset((cls.paid, cls.rejected))

    @classproperty
    def pending(cls):
        return frozenset((cls.evaluating, cls.approved, cls.incomplete))

    @classproperty
    def statuses(cls):
        return frozenset((cls.evaluating, cls.approved, cls.paid, cls.rejected,
                cls.incomplete))


class ActionError(ValueError):
    """Error raised for invalid state changes for a :py:class:`Request`."""
    pass


class Action(db.Model, AutoID, Timestamped, AutoName):
    """Actions change the state of a Request.
    
    :py:class:`Request`\s enforce permissions when actions are added to them.
    If the user adding the action does not have the appropriate
    :py:class:`~.Permission`\s in the request's :py:class:`Division`, an
    :py:exc:`ActionError` will be raised.

    With the exception of the :py:attr:`comment <ActionType.comment>` action
    (which just adds text to a request), actions change the
    :py:attr:`~Request.status` of a Request.
    """

    #: The action be taken. See :py:class:`ActionType` for possible values.
    # See set_request_type below for the effect setting this attribute has on
    # the parent Request.
    type_ = db.Column(ActionType.db_type(), nullable=False)

    #: The ID of the :py:class:`Request` this action applies to.
    request_id = db.Column(db.Integer, db.ForeignKey('request.id'))

    #: The :py:class:`Request` this action applies to.
    request = db.relationship('Request', back_populates='actions',
            cascade='save-update,merge,refresh-expire,expunge')

    #: The ID of the :py:class:`~.User` who made this action.
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)

    #: The :py:class:`~.User` who made this action.
    user = db.relationship('User', back_populates='actions',
            cascade='save-update,merge,refresh-expire,expunge')

    #: Any additional notes for this action.
    note = db.Column(db.Text(convert_unicode=True))

    def __init__(self, request, user, note=None, type_=None):
        if type_ is not None:
            self.type_ = type_
        self.user = user
        self.note = ensure_unicode(note)
        # timestamp has to be an actual value (besides None) before the request
        # is set so thhe request's validation doesn't fail.
        self.timestamp = dt.datetime.utcnow()
        self.request = request

    def __repr__(self):
        return "{x.__class__.__name__}({x.request}, {x.user}, {x.type_})".\
                format(x=self)

    def _json(self, extended=False):
        try:
            parent = super(Action, self)._json(extended)
        except AttributeError:
            parent = {}
        parent[u'type'] = self.type_
        if extended:
            parent[u'note'] = self.note or u''
            parent[u'timestamp'] = self.timestamp
            parent[u'user'] = self.user
        return parent


class ModifierError(ValueError):
    """Error raised when a modification is attempted to a :py:class:`Request`
    when it's in an invalid state.
    """
    pass


class Modifier(db.Model, AutoID, Timestamped, AutoName):
    """Modifiers apply bonuses or penalties to Requests.

    This is an abstract base class for the pair of concrete implementations.
    Modifiers can be voided at a later date. The user who voided a modifier and
    when it was voided are recorded.

    :py:class:`Request`\s enforce permissions when modifiers are added. If the
    user adding a modifier does not have the appropriate
    :py:class:`~.Permission`\s in the request's :py:class:`~.Division`, a
    :py:exc:`ModifierError` will be raised.
    """

    #: Discriminator column for SQLAlchemy
    _type = db.Column(db.String(20, convert_unicode=True), nullable=False)

    #: The ID of the :py:class:`Request` this modifier applies to.
    request_id = db.Column(db.Integer, db.ForeignKey('request.id'))

    #: The :py:class:`Request` this modifier applies to.
    request = db.relationship('Request', back_populates='modifiers',
            cascade='save-update,merge,refresh-expire,expunge')

    #: The ID of the :py:class`~.User` who added this modifier.
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)

    #: The :py:class:`~.User` who added this modifier.
    user = db.relationship('User', foreign_keys=[user_id],
            cascade='save-update,merge,refresh-expire,expunge')

    #: Any notes explaining this modification.
    note = db.Column(db.Text(convert_unicode=True))

    #: The ID of the :py:class:`~.User` who voided this modifier (if voided).
    voided_user_id = db.Column(db.Integer, db.ForeignKey('user.id'),
            nullable=True)

    #: The :py:class:`~.User` who voided this modifier if it has been voided.
    voided_user = db.relationship('User', foreign_keys=[voided_user_id],
            cascade='save-update,merge,refresh-expire,expunge')

    #: If this modifier has been voided, this will be the timestamp of when it
    #: was voided.
    voided_timestamp = db.Column(DateTime)

    @hybrid_property
    def voided(self):
        return self.voided_user is not None and \
                self.voided_timestamp is not None

    @classmethod
    def _voided_select(cls):
        """Create a subquery with two columns, ``modifier_id`` and ``voided``.

        Used for the expressions of :py:attr:`voided` and
        :py:attr:`Request.payout`.
        """
        user = db.select([cls.id.label('modifier_id'),
                cls.voided_user_id.label('user_id')]).alias('user_sub')
        timestamp = db.select([cls.id.label('modifier_id'),
                cls.voided_timestamp.label('timestamp')]).alias('timestamp_sub')
        columns = [
            db.and_(
                user.c.user_id != None,
                timestamp.c.timestamp != None).label('voided'),
            user.c.modifier_id.label('modifier_id'),
        ]
        return db.select(columns).where(
                user.c.modifier_id == timestamp.c.modifier_id)\
                .alias('voided_sub')

    @voided.expression
    def voided(cls):
        return cls._voided_select().c.voided

    @declared_attr
    def __mapper_args__(cls):
        """SQLAlchemy late-binding attribute to set mapper arguments.

        Obviates subclasses from having to specify polymorphic identities.
        """
        cls_name = unicode(cls.__name__)
        args = {'polymorphic_identity': cls_name}
        if cls_name == u'Modifier':
            args['polymorphic_on'] = cls._type
        return args

    def __init__(self, request, user, note, value):
        self.user = user
        self.note = ensure_unicode(note)
        self.value = value
        self.request = request

    def __repr__(self):
        return ("{x.__class__.__name__}({x.request}, {x.user},"
                "{x}, {x.voided})".format(x=self, value=self))

    def void(self, user):
        """Mark this modifier as void.

        :param user: The user voiding this modifier
        :type user: :py:class:`~.User`
        """
        if self.request.status != ActionType.evaluating:
            # TRANS: Error message shown when trying to void (cancel) a
            # modifier but the request is not in the evaluating state, so the
            # attempt fails.
            raise ModifierError(gettext(u"Modifiers can only be voided when "
                                        u"the request is in the evaluating "
                                        u"state."))
        if not user.has_permission(PermissionType.review,
                self.request.division):
            # TRANS: Error message shown when you attempt to void a modifier
            # but are prevented from doing so because you do not hold the
            # reviewer permission.
            raise ModifierError("You must be a reviewer to be able to void "
                                "modifiers.")
        self.voided_user = user
        self.voided_timestamp = dt.datetime.utcnow()

    @db.validates('request')
    def _check_request_status(self, attr, request):
        if current_app.config['SRP_SKIP_VALIDATION']:
            return request
        if request.status != ActionType.evaluating:
            raise ModifierError(gettext(u"Modifiers can only be added when the"
                                        u" request is in an evaluating "
                                        u"state."))
        if not self.user.has_permission(PermissionType.review,
                request.division):
            raise ModifierError(gettext(u"Only reviewers can add modifiers."))
        return request

    def _json(self, extended=False):
        try:
            parent = super(Modifier, self)._json(extended)
        except AttributeError:
            parent = {}
        parent[u'value'] = self.value
        if extended:
            parent[u'note'] = self.note or u''
            parent[u'timestamp'] = self.timestamp
            parent[u'user'] = self.user
            if self.voided:
                parent[u'void'] = {
                    u'user': self.voided_user,
                    u'timestamp': self.voided_timestamp,
                }
            else:
                parent[u'void'] = False
        else:
            parent[u'void'] = self.voided
        return parent


@unistr
class AbsoluteModifier(Modifier):
    """Subclass of :py:class:`Modifier` for representing absolute
    modifications.

    Absolute modifications are those that are not dependent on the value of
    :py:attr:`Request.base_payout`.
    """

    id = db.Column(db.Integer, db.ForeignKey('modifier.id'), primary_key=True)

    #: How much ISK to add or remove from the payout
    value = db.Column(PrettyNumeric(precision=15, scale=2), nullable=False,
            default=Decimal(0))

    def _json(self, extended=False):
        try:
            parent = super(AbsoluteModifier, self)._json(extended)
        except AttributeError:
            parent = {}
        parent[u'type'] = 'absolute'
        return parent


@unistr
class RelativeModifier(Modifier):
    """Subclass of :py:class:`Modifier` for representing relative modifiers.

    Relative modifiers depend on the value of :py:attr:`Modifier.base_payout`
    to calculate their effect.
    """

    id = db.Column(db.Integer, db.ForeignKey('modifier.id'), primary_key=True)

    #: What percentage of the payout to add or remove
    value = db.Column(db.Numeric(precision=8, scale=5), nullable=False,
            default=Decimal(0))

    def _json(self, extended=False):
        try:
            parent = super(RelativeModifier, self)._json(extended)
        except AttributeError:
            parent = {}
        parent[u'type'] = 'relative'
        return parent


class Request(db.Model, AutoID, Timestamped, AutoName):
    """Requests represent SRP requests."""

    #: The ID of the :py:class:`~.User` who submitted this request.
    submitter_id = db.Column(db.Integer, db.ForeignKey('user.id'))

    #: The :py:class:`~.User` who submitted this request.
    submitter = db.relationship('User', back_populates='requests',
            cascade='save-update,merge,refresh-expire,expunge')

    #: The ID of the :py:class`~.Division` this request was submitted to.
    division_id = db.Column(db.Integer, db.ForeignKey('division.id'),
            nullable=False)

    #: The :py:class:`~.Division` this request was submitted to.
    division = db.relationship('Division', back_populates='requests',
            cascade='save-update,merge,refresh-expire,expunge')

    #: A list of :py:class:`Action`\s that have been applied to this request,
    #: sorted in the order they were applied.
    actions = db.relationship('Action', back_populates='request',
            cascade='all,delete-orphan',
            order_by='desc(Action.timestamp)')

    #: A list of all :py:class:`Modifier`\s that have been applied to this
    #: request, regardless of wether they have been voided or not. They're
    #: sorted in the order they were added.
    modifiers = db.relationship('Modifier', back_populates='request',
            cascade='all,delete-orphan',
            lazy='dynamic', order_by='desc(Modifier.timestamp)')

    #: The URL of the source killmail.
    killmail_url = db.Column(db.String(512, convert_unicode=True),
            nullable=False)

    #: The ID of the :py:class:`~.Pilot` for the killmail.
    pilot_id = db.Column(db.Integer, db.ForeignKey('pilot.id'), nullable=False)

    #: The :py:class:`~.Pilot` who was the victim in the killmail.
    pilot = db.relationship('Pilot', back_populates='requests',
            cascade='save-update,merge,refresh-expire,expunge')

    #: The corporation of the :py:attr:`pilot` at the time of the killmail.
    corporation = db.Column(db.String(150, convert_unicode=True),
            nullable=False, index=True)

    #: The alliance of the :py:attr:`pilot` at the time of the killmail.
    alliance = db.Column(db.String(150, convert_unicode=True), nullable=True,
            index=True)

    #: The type of ship that was destroyed.
    ship_type = db.Column(db.String(75, convert_unicode=True), nullable=False,
            index=True)

    # TODO: include timezones
    #: The date and time of when the ship was destroyed.
    kill_timestamp = db.Column(DateTime, nullable=False, index=True)

    base_payout = db.Column(PrettyNumeric(precision=15, scale=2),
            default=Decimal(0))
    """The base payout for this request.

    This value is clamped to a lower limit of 0. It can only be changed when
    this request is in an :py:attr:`~ActionType.evaluating` state, or else a
    :py:exc:`ModifierError` will be raised.
    """

    #: The payout for this requests taking into account all active modifiers.
    payout = db.Column(PrettyNumeric(precision=15, scale=2),
            default=Decimal(0), index=True, nullable=False)

    #: Supporting information for the request.
    details = db.deferred(db.Column(db.Text(convert_unicode=True)))

    #: The current status of this request
    status = db.Column(ActionType.db_type(), nullable=False,
            default=ActionType.evaluating)
    """This attribute is automatically kept in sync as :py:class:`Action`\s are
    added to the request. It should not be set otherwise.

    At the time an :py:class:`Action` is added to this request, the type of
    action is checked and the state diagram below is enforced. If the action is
    invalid, an :py:exc:`ActionError` is raised.

    .. digraph:: request_workflow

        rankdir="LR";

        sub [label="submitted", shape=plaintext];

        node [style="dashed, filled"];

        eval [label="evaluating", fillcolor="#fcf8e3"];
        rej [label="rejected", style="solid, filled", fillcolor="#f2dede"];
        app [label="approved", fillcolor="#d9edf7"];
        inc [label="incomplete", fillcolor="#f2dede"];
        paid [label="paid", style="solid, filled", fillcolor="#dff0d8"];

        sub -> eval;
        eval -> rej [label="R"];
        eval -> app [label="R"];
        eval -> inc [label="R"];
        rej -> eval [label="R"];
        inc -> eval [label="R, S"];
        inc -> rej [label="R"];
        app -> paid [label="P"];
        app -> eval [label="R"];
        paid -> eval [label="P"];
        paid -> app [label="P"];

    R means a reviewer can make that change, S means the submitter can make
    that change, and P means payers can make that change. Solid borders are
    terminal states.
    """

    #: The solar system this loss occured in.
    system = db.Column(db.String(25, convert_unicode=True), nullable=False,
            index=True)

    #: The constellation this loss occured in.
    constellation = db.Column(db.String(25, convert_unicode=True),
            nullable=False, index=True)

    #: The region this loss occured in.
    region = db.Column(db.String(25, convert_unicode=True), nullable=False,
            index=True)

    @hybrid_property
    def finalized(self):
        return self.status in ActionType.finalized

    @finalized.expression
    def finalized(cls):
        return db.or_(cls.status == ActionType.paid,
                cls.status == ActionType.rejected)

    def __init__(self, submitter, details, division, killmail, **kwargs):
        """Create a :py:class:`Request`.

        :param submitter: The user submitting this request
        :type submitter: :py:class:`~.User`
        :param str details: Supporting details for this request
        :param division: The division this request is being submitted to
        :type division: :py:class:`~.Division`
        :param killmail: The killmail this request pertains to
        :type killmail: :py:class:`~.Killmail`
        """
        with db.session.no_autoflush:
            self.division = division
            self.details = details
            self.submitter = submitter
            # Pull basically everything else from the killmail object
            # The base Killmail object has an iterator defined that returns tuples
            # of Request attributes and values for those attributes
            for attr, value in killmail:
                setattr(self, attr, value)
            # Set default values before a flush
            if self.base_payout is None and 'base_payout' not in kwargs:
                self.base_payout = Decimal(0)
        super(Request, self).__init__(**kwargs)

    @db.validates('base_payout')
    def _validate_payout(self, attr, value):
        """Ensures that base_payout is positive. The value is clamped to 0."""
        if current_app.config['SRP_SKIP_VALIDATION']:
            return Decimal(value)
        # Allow self.status == None, as the base payout may be set in the
        # initializing state before the status has been set.
        if self.status == ActionType.evaluating or self.status is None:
            if value is None or value < 0:
                return Decimal('0')
            else:
                return Decimal(value)
        else:
            raise ModifierError(gettext(u"The request must be in the "
                                        u"evaluating state to change the base "
                                        u"payout."))

    state_rules = {
        ActionType.evaluating: {
            ActionType.incomplete: (PermissionType.review,
                PermissionType.admin),
            ActionType.rejected: (PermissionType.review,
                PermissionType.admin),
            ActionType.approved: (PermissionType.review,
                PermissionType.admin),
        },
        ActionType.incomplete: {
            ActionType.rejected: (PermissionType.review,
                PermissionType.admin),
            # Special case: the submitter can change it to evaluating by
            # changing the division or updating the details.
            ActionType.evaluating: (PermissionType.review,
                PermissionType.admin),
        },
        ActionType.rejected: {
            ActionType.evaluating: (PermissionType.review,
                PermissionType.admin),
        },
        ActionType.approved: {
            # Special case: the submitter can change it to evaluating by
            # changing the division.
            ActionType.evaluating: (PermissionType.review,
                PermissionType.admin),
            ActionType.paid: (PermissionType.pay, PermissionType.admin),
        },
        ActionType.paid: {
            ActionType.approved: (PermissionType.pay, PermissionType.admin),
            ActionType.evaluating: (PermissionType.pay, PermissionType.admin),
        },
    }

    def valid_actions(self, user):
        """Get valid actions (besides comment) the given user can perform."""
        possible_actions = self.state_rules[self.status]
        def action_filter(action):
            return user.has_permission(possible_actions[action],
                    self.division)
        return filter(action_filter, possible_actions)

    @db.validates('status')
    def _validate_status(self, attr, new_status):
        """Enforces that status changes follow the status state machine.
        When an invalid change is attempted, :py:class:`ActionError` is
        raised.
        """
        if current_app.config['SRP_SKIP_VALIDATION']:
            return new_status
        if new_status == ActionType.comment:
            raise ValueError(gettext(
                u"Comment is not a valid status"))
        # Initial status
        if self.status is None:
            return new_status
        rules = self.state_rules[self.status]
        if new_status not in rules:
            error_text = gettext(u"%(new_status)s is not a valid status to "
                                 u"change to from %(old_status)s.",
                    new_status=new_status,
                    old_status=self.status)
            raise ActionError(error_text)
        return new_status

    @db.validates('actions')
    def _verify_action_permissions(self, attr, action):
        """Verifies that permissions for Actions being added to a Request."""
        if current_app.config['SRP_SKIP_VALIDATION']:
            return action
        if action.type_ is None:
            # Action.type_ are not nullable, so rely on the fact that it will
            # be set later to let it slide now.
            return action
        elif action.type_ != ActionType.comment:
            # Peek behind the curtain to see the history of the status
            # attribute.
            status_history = db.inspect(self).attrs.status.history
            if status_history.has_changes():
                new_status = status_history.added[0]
                old_status = status_history.deleted[0]
            else:
                new_status = action.type_
                old_status = self.status
            rules = self.state_rules[old_status]
            permissions = rules[new_status]
            # Handle the special cases called out in state_rules
            if action.user == self.submitter and \
                    new_status == ActionType.evaluating and \
                    old_status in ActionType.pending:
                # Equivalent to self.status in (approved, incomplete) as
                # going from evaluating to evaluating is invalid (as checked by
                # the status validator).
                return action
            if not action.user.has_permission(permissions, self.division):
                raise ActionError(gettext(u"Insufficient permissions to "
                                          u"perform that action."))
        elif action.type_ == ActionType.comment:
            if action.user != self.submitter \
                    and not action.user.has_permission(
                            (PermissionType.review, PermissionType.pay,
                                PermissionType.admin),
                            self.division):
                raise ActionError(gettext(u"You must either own or have "
                                          u"special privileges to comment on "
                                          u"this request."))
        return action

    def __repr__(self):
        return "{x.__class__.__name__}({x.submitter}, {x.division}, {x.id})".\
                format(x=self)

    @property
    def transformed(self):
        """Get a special HTML representation of an attribute.

        Divisions can have a transformer defined on various attributes that
        output a URL associated with that attribute. This property provides
        easy access to the output of any transformed attributes on this
        request.
        """
        class RequestTransformer(object):
            def __init__(self, request):
                self._request = request

            def __getattr__(self, attr):
                raw_value = getattr(self._request, attr)
                if attr in self._request.division.transformers:
                    transformer = self._request.division.transformers[attr]
                    return Markup(u'<a href="{link}" target="_blank">'
                                  u'{value} <i class="fa fa-external-link">'
                                  u'</i></a>').format(
                                        link=transformer(raw_value),
                                        value=str(raw_value))
                else:
                    return raw_value

            def __iter__(self):
                for attr, transformer in\
                        self._request.division.transformers.items():
                    if attr == 'ship_type':
                        yield ('ship', transformer(getattr(self._request,
                                attr)))
                    else:
                        yield (attr, transformer(getattr(self._request, attr)))

        return RequestTransformer(self)

    def _json(self, extended=False):
        try:
            parent = super(Request, self)._json(extended)
        except AttributeError:
            parent = {}
        parent[u'href'] = url_for('requests.get_request_details',
                request_id=self.id)
        attrs = (u'killmail_url', u'kill_timestamp', u'pilot',
                 u'alliance', u'corporation', u'submitter',
                 u'division', u'status', u'base_payout', u'payout',
                 u'details', u'id', u'ship_type', u'system', u'constellation',
                 u'region')
        for attr in attrs:
            if attr == u'ship_type':
                parent['ship'] = self.ship_type
            elif u'payout' in attr:
                payout = getattr(self, attr)
                parent[attr] = payout.currency()
            else:
                parent[attr] = getattr(self, attr)
        parent[u'submit_timestamp'] = self.timestamp
        if extended:
            parent[u'actions'] = map(lambda a: a._json(True), self.actions)
            parent[u'modifiers'] = map(lambda m: m._json(True), self.modifiers)
            parent[u'valid_actions'] = self.valid_actions(current_user)
            parent[u'transformed'] = dict(self.transformed)
        return parent


# Define event listeners for syncing the various denormalized attributes

@listens_for(Action.type_, 'set')
def _action_type_to_request_status(action, new_status, old_status, initiator):
    """Set the Action's Request's status when the Action's type is changed."""
    if action.request is not None and new_status != ActionType.comment:
        action.request.status = new_status


@listens_for(Request.actions, 'append')
def _request_status_from_actions(srp_request, action, initiator):
    """Updates Request.status when new Actions are added."""
    # Pass when Action.type_ is None, as it'll get updated later
    if action.type_ is not None and action.type_ != ActionType.comment:
        srp_request.status = action.type_


@listens_for(Request.base_payout, 'set')
def _recalculate_payout_from_request(srp_request, base_payout, *args):
    """Recalculate a Request's payout when the base payout changes."""
    if base_payout is None:
        base_payout = Decimal(0)
    voided = Modifier._voided_select()
    modifiers = srp_request.modifiers.join(voided,
                voided.c.modifier_id==Modifier.id)\
            .filter(~voided.c.voided)\
            .order_by(False)
    absolute = modifiers.join(AbsoluteModifier).\
            with_entities(db.func.sum(AbsoluteModifier.value)).\
            scalar()
    if not isinstance(absolute, Decimal):
        absolute = Decimal(0)
    relative = modifiers.join(RelativeModifier).\
            with_entities(db.func.sum(RelativeModifier.value)).\
            scalar()
    if not isinstance(relative, Decimal):
        relative = Decimal(0)
    payout = (base_payout + absolute) * (Decimal(1) + relative)
    srp_request.payout = PrettyDecimal(payout)


@listens_for(Modifier.request, 'set', propagate=True)
@listens_for(Modifier.voided_user, 'set', propagate=True)
def _recalculate_payout_from_modifier(modifier, value, *args):
    """Recalculate a Request's payout when it gains a Modifier or when one of
    its Modifiers is voided.
    """
    # Force a flush at the beginning, then delay other flushes
    db.session.flush()
    with db.session.no_autoflush:
        # Get the request for this modifier
        if isinstance(value, Request):
            # Triggered by setting Modifier.request
            srp_request = value
        else:
            # Triggered by setting Modifier.voided_user
            srp_request = modifier.request
        voided = Modifier._voided_select()
        modifiers = srp_request.modifiers.join(voided,
                    voided.c.modifier_id==Modifier.id)\
                .filter(~voided.c.voided)\
                .order_by(False)
        absolute = modifiers.join(AbsoluteModifier).\
                with_entities(db.func.sum(AbsoluteModifier.value)).\
                scalar()
        if not isinstance(absolute, Decimal):
            absolute = Decimal(0)
        relative = modifiers.join(RelativeModifier).\
                with_entities(db.func.sum(RelativeModifier.value)).\
                scalar()
        if not isinstance(relative, Decimal):
            relative = Decimal(0)
        # The modifier that's changed isn't reflected yet in the database, so we
        # apply it here.
        if isinstance(value, Request):
            # A modifier being added to the Request
            if modifier.voided:
                # The modifier being added is already void
                return
            direction = Decimal(1)
        else:
            # A modifier already on a request is being voided
            direction = Decimal(-1)
        if isinstance(modifier, AbsoluteModifier):
            absolute += direction * modifier.value
        elif isinstance(modifier, RelativeModifier):
            relative += direction * modifier.value
        payout = (srp_request.base_payout + absolute) * \
                (Decimal(1) + relative)
        srp_request.payout = PrettyDecimal(payout)


# The next few lines are responsible for adding a full text search index on the
# Request.details column for MySQL.
_create_fts = DDL('CREATE FULLTEXT INDEX ix_%(table)s_details_fulltext '
                       'ON %(table)s (details);')
_drop_fts = DDL('DROP INDEX ix_%(table)s_details_fulltext ON %(table)s')


event.listen(
        Request.__table__,
        'after_create',
        _create_fts.execute_if(dialect='mysql')
)


event.listen(
        Request.__table__,
        'before_drop',
        _drop_fts.execute_if(dialect='mysql')
)



from __future__ import absolute_import

import wtforms
import wtforms.widgets
import wtforms.fields
from wtforms.utils import unset_value


class ImageInput(wtforms.widgets.Input):
    """WTForms widget for image inputs (<input type="image">)
    """

    input_type = u'image'

    def __init__(self, src='', alt=''):
        super(ImageInput, self).__init__()
        self.src = src
        self.alt = alt

    def __call__(self, field, **kwargs):
        kwargs['src'] = self.src
        kwargs['alt'] = self.alt
        return super(ImageInput, self).__call__(field, **kwargs)


class ImageField(wtforms.fields.BooleanField):
    """WTForms field for image fields.
    """

    def __init__(self, src, alt='', **kwargs):
        widget = ImageInput(src, alt)
        super(wtforms.fields.BooleanField, self).__init__(widget=widget,
                **kwargs)

    def process(self, formdata, data=unset_value):
        if formdata:
            for key in formdata:
                if key.startswith(self.name):
                    self.data = True
                    break
            else:
                self.data = False
        else:
            self.data = False



from __future__ import unicode_literals

from .util.crest import NameLookup
from . import static_data


ships = NameLookup(static_data.ships,
        'itemTypes',
        'application/vnd.ccp.eve.ItemType-v3+json')



"""Add an auditor-level permission, and fix Enum item order

Revision ID: 8f179e969c9
Revises: 5795c29b2c7a
Create Date: 2014-07-24 17:24:12.026685

"""

# revision identifiers, used by Alembic.
revision = '8f179e969c9'
down_revision = '3e5e1d3a02c'

from alembic import op
import sqlalchemy as sa


new_enum = sa.Enum(u'admin', u'audit', u'review', u'pay', u'submit',
        convert_unicode=True,
        name=u'ck_permission_type')


old_enum = sa.Enum(u'admin', u'review', u'pay', u'submit',
        convert_unicode=True,
        name=u'ck_permission_type')


def upgrade():
    op.alter_column('permission', 'permission', type_=new_enum,
            existing_type=old_enum, existing_nullable=False)


def downgrade():
    op.alter_column('permission', 'permission', type_=old_enum,
            existing_type=new_enum, existing_nullable=False)



"""Add location attributes to Request.

Revision ID: 2f22504b1e6
Revises: None
Create Date: 2014-05-02 13:24:00.045482

"""

# revision identifiers, used by Alembic.
revision = '2f22504b1e6'
down_revision = None

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import update, select, table, column
import requests
from time import sleep
from evesrp import systems


session = requests.Session()


def get_system_id(kill_id):
    # Have to put a sleep in here to prevent hammering the server and getting
    # banned
    sleep(10)
    resp = session.get(
            'https://zkb.pleaseignore.com/api/no-attackers/no-items/killID/{}'
            .format(kill_id))
    if resp.status_code == 200:
        return int(resp.json()[0]['solarSystemID'])
    else:
        print(resp.text)
        return None


request = table('request',
        column('id', sa.Integer),
        column('system', sa.String(25)),
        column('constellation', sa.String(25)),
        column('region', sa.String(25)),
)


def upgrade():
    # Add columns, but with null allowed
    op.add_column('request', sa.Column('constellation', sa.String(length=25),
            nullable=True))
    op.add_column('request', sa.Column('region', sa.String(length=25),
            nullable=True))
    op.add_column('request', sa.Column('system', sa.String(length=25),
            nullable=True))
    op.create_index('ix_request_constellation', 'request', ['constellation'],
            unique=False)
    op.create_index('ix_request_region', 'request', ['region'], unique=False)
    op.create_index('ix_request_system', 'request', ['system'], unique=False)
    # Update existing requests
    conn = op.get_bind()
    kill_id_sel = select([request.c.id])
    kill_ids = conn.execute(kill_id_sel)
    for kill_id in kill_ids:
        kill_id = kill_id[0]
        system_id = get_system_id(kill_id)
        system = systems.system_names[system_id]
        constellation = systems.systems_constellations[system]
        region = systems.constellations_regions[constellation]
        update_stmt = update(request)\
                .where(request.c.id==op.inline_literal(kill_id))\
                .values({
                    'system': system,
                    'constellation': constellation,
                    'region': region,
                })
        conn.execute(update_stmt)
    kill_ids.close()
    # Add non-null constraint
    op.alter_column('request', 'constellation', nullable=False,
            existing_server_default=None,
            existing_type=sa.String(length=25))
    op.alter_column('request', 'region', nullable=False,
            existing_server_default=None,
            existing_type=sa.String(length=25))
    op.alter_column('request', 'system', nullable=False,
            existing_server_default=None,
            existing_type=sa.String(length=25))



def downgrade():
    op.drop_index('ix_request_system', table_name='request')
    op.drop_index('ix_request_region', table_name='request')
    op.drop_index('ix_request_constellation', table_name='request')
    op.drop_column('request', 'system')
    op.drop_column('request', 'region')
    op.drop_column('request', 'constellation')



from __future__ import absolute_import

import datetime
import ssl
import sys
import warnings

from requests.adapters import HTTPAdapter
from requests.packages import urllib3
from requests.packages.urllib3.util import ssl_

from requests.packages.urllib3.exceptions import (
    SystemTimeWarning,
    SecurityWarning,
)
from requests.packages.urllib3.packages.ssl_match_hostname import \
        match_hostname


class WeakCiphersHTTPSConnection(
        urllib3.connection.VerifiedHTTPSConnection): # pragma: no cover

    # Python versions >=2.7.9 and >=3.4.1 do not (by default) allow ciphers
    # with MD5. Unfortunately, the CREST public server _only_ supports
    # TLS_RSA_WITH_RC4_128_MD5 (as of 5 Jan 2015). The cipher list below is
    # nearly identical except for allowing that cipher as a last resort (and
    # excluding export versions of ciphers).
    DEFAULT_CIPHERS = (
        'ECDH+AESGCM:DH+AESGCM:ECDH+AES256:DH+AES256:ECDH+AES128:DH+AES:'
        'ECDH+HIGH:DH+HIGH:ECDH+3DES:DH+3DES:RSA+AESGCM:RSA+AES:RSA+HIGH:'
        'RSA+3DES:ECDH+RC4:DH+RC4:RSA+RC4:!aNULL:!eNULL:!EXP:-MD5:RSA+RC4+MD5'
    )

    def __init__(self, host, port, ciphers=None, **kwargs):
        self.ciphers = ciphers if ciphers is not None else self.DEFAULT_CIPHERS
        super(WeakCiphersHTTPSConnection, self).__init__(host, port, **kwargs)

    def connect(self):
        # Yup, copied in VerifiedHTTPSConnection.connect just to change the
        # default cipher list.

        # Add certificate verification
        conn = self._new_conn()

        resolved_cert_reqs = ssl_.resolve_cert_reqs(self.cert_reqs)
        resolved_ssl_version = ssl_.resolve_ssl_version(self.ssl_version)

        hostname = self.host
        if getattr(self, '_tunnel_host', None):
            # _tunnel_host was added in Python 2.6.3
            # (See: http://hg.python.org/cpython/rev/0f57b30a152f)

            self.sock = conn
            # Calls self._set_hostport(), so self.host is
            # self._tunnel_host below.
            self._tunnel()
            # Mark this connection as not reusable
            self.auto_open = 0

            # Override the host with the one we're requesting data from.
            hostname = self._tunnel_host

        is_time_off = datetime.date.today() < urllib3.connection.RECENT_DATE
        if is_time_off:
            warnings.warn((
                'System time is way off (before {0}). This will probably '
                'lead to SSL verification errors').format(
                    urllib3.connection.RECENT_DATE),
                SystemTimeWarning
            )

        # Wrap socket using verification with the root certs in
        # trusted_root_certs
        self.sock = ssl_.ssl_wrap_socket(conn, self.key_file, self.cert_file,
                                    cert_reqs=resolved_cert_reqs,
                                    ca_certs=self.ca_certs,
                                    server_hostname=hostname,
                                    ssl_version=resolved_ssl_version,
                                    ciphers=self.ciphers)

        if self.assert_fingerprint:
            ssl_.assert_fingerprint(self.sock.getpeercert(binary_form=True),
                               self.assert_fingerprint)
        elif resolved_cert_reqs != ssl.CERT_NONE \
                and self.assert_hostname is not False:
            cert = self.sock.getpeercert()
            if not cert.get('subjectAltName', ()):
                warnings.warn((
                    'Certificate has no `subjectAltName`, falling back to check for a `commonName` for now. '
                    'This feature is being removed by major browsers and deprecated by RFC 2818. '
                    '(See https://github.com/shazow/urllib3/issues/497 for details.)'),
                    SecurityWarning
                )
            match_hostname(cert, self.assert_hostname or hostname)

        self.is_verified = (resolved_cert_reqs == ssl.CERT_REQUIRED
                            or self.assert_fingerprint is not None)


class WeakCiphersHTTPSConnectionPool(
        urllib3.connectionpool.HTTPSConnectionPool):

    ConnectionCls = WeakCiphersHTTPSConnection


class WeakCiphersPoolManager(urllib3.poolmanager.PoolManager):

    def _new_pool(self, scheme, host, port):
        if scheme == 'https':
            return WeakCiphersHTTPSConnectionPool(host, port,
                    **(self.connection_pool_kw))
        return super(WeakCiphersPoolManager, self)._new_pool(scheme, host,
                port)


class WeakCiphersAdapter(HTTPAdapter):
    """"Transport adapter" that allows us to use TLS_RSA_WITH_RC4_128_MD5."""

    def init_poolmanager(self, connections, maxsize, block=False,
            **pool_kwargs):
        # Rewrite of the requests.adapters.HTTPAdapter.init_poolmanager method
        # to use WeakCiphersPoolManager instead of urllib3's PoolManager
        self._pool_connections = connections
        self._pool_maxsize = maxsize
        self._pool_block = block

        self.poolmanager = WeakCiphersPoolManager(num_pools=connections,
                maxsize=maxsize, block=block, strict=True, **pool_kwargs)



from functools import wraps
from flask import make_response
import six


def varies(*on):
    def vary_decorator(func):
        @wraps(func)
        def vary_decorated(*args, **kwargs):
            response = make_response(func(*args, **kwargs))
            response.headers['Vary'] = ', '.join(on)
            return response
        return vary_decorated
    return vary_decorator



import re
from flask import current_app
from .util.crest import check_crest_response, NameLookup
from . import static_data


REGION_TYPE = 'application/vnd.ccp.eve.Region-v1+json'
CONSTELLATION_TYPE = 'application/vnd.ccp.eve.Constellation-v1+json'
SYSTEM_TYPE = 'application/vnd.ccp.eve.System-v1+json'


region_names = NameLookup(static_data.region_names, 'regions', REGION_TYPE)


constellation_names = NameLookup(
        static_data.constellation_names,
        'constellations',
        CONSTELLATION_TYPE)


system_names = NameLookup(static_data.system_names, 'systems', SYSTEM_TYPE)


class ConstellationRegionLookup(NameLookup):

    def __init__(self):
        super(ConstellationRegionLookup, self).__init__(
                static_data.constellations_to_regions,
                'constellations',
                CONSTELLATION_TYPE,
                'region.href')

    def __getitem__(self, key):
        parent_item = super(ConstellationRegionLookup, self).__getitem__(key)
        # parent item will either be a CREST URL or an integer type ID
        if isinstance(parent_item, int):
            return parent_item
        resp = current_app.requests_session.get(parent_item,
                headers={'Accept': REGION_TYPE})
        if check_crest_response(resp) and resp.status_code == 200:
            self._dict[key] = resp.json()['id']
        else:
            message = "Cannot find the name for the ID requested: {}"\
                    .format(key)
            raise KeyError(message)
        return self._dict[key]


systems_constellations = NameLookup(static_data.systems_to_constellations,
                                    'systems',
                                    SYSTEM_TYPE,
                                    'constellation.id')

constellations_regions = ConstellationRegionLookup()



"""Change RelativeModifier.value to a Numeric/Decimal

Revision ID: 5795c29b2c7a
Revises: 19506187e7aa
Create Date: 2014-07-23 14:43:45.748696

"""

# revision identifiers, used by Alembic.
revision = '5795c29b2c7a'
down_revision = '19506187e7aa'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import update, select, table, column
from decimal import Decimal


def upgrade():
    relative_modifier = table('relative_modifier',
            column('id', sa.Integer),
            column('value', sa.Float),
            column('numeric_value', sa.Numeric(precision=8, scale=5)))
    op.add_column('relative_modifier',
            sa.Column('numeric_value', sa.Numeric(precision=8, scale=5)))
    conn = op.get_bind()
    sel = select([relative_modifier.c.id, relative_modifier.c.value])
    results = conn.execute(sel)
    q = Decimal(10) ** -5
    for id_, float_value in results:
        decimal_value = Decimal(float_value).quantize(q)
        up = update(relative_modifier).where(relative_modifier.c.id == id_)\
                .values({'numeric_value': decimal_value})
        conn.execute(up)
    op.drop_column('relative_modifier', 'value')
    op.alter_column('relative_modifier', 'numeric_value', nullable=True,
            new_column_name='value', existing_type=sa.Numeric(precision=8,
                    scale=5))


def downgrade():
    relative_modifier = table('relative_modifier',
            column('id', sa.Integer),
            column('value', sa.Numeric(precision=8, scale=5)),
            column('float_value', sa.Float))
    op.add_column('relative_modifier', sa.Column('float_value', sa.Float))
    conn = op.get_bind()
    sel = select([relative_modifier.c.id, relative_modifier.c.value])
    results = conn.execute(sel)
    for id_, decimal_value in results:
        float_value = float(decimal_value)
        up = update(relative_modifier).where(relative_modifier.c.id == id_)\
                .values({'float_value': float_value})
        conn.execute(up)
    op.drop_column('relative_modifier', 'value')
    op.alter_column('relative_modifier', 'float_value', nullable=True,
            new_column_name='value', existing_type=sa.Float)



"""Denormalize Request.payout

Revision ID: 3e5e1d3a02c
Revises: 5795c29b2c7a
Create Date: 2014-07-26 21:25:29.870535

"""

# revision identifiers, used by Alembic.
revision = '3e5e1d3a02c'
down_revision = '5795c29b2c7a'

from decimal import Decimal
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import update, select, table, column, join, outerjoin, case
from sqlalchemy.sql.functions import func


request = table('request',
        column('id', sa.Integer),
        column('base_payout', sa.Numeric(precision=15, scale=2)),
        column('payout', sa.Numeric(precision=15, scale=2)))


mod_table = table('modifier',
        column('id', sa.Integer),
        column('request_id', sa.Integer),
        column('_type', sa.String(length=20)),
        column('voided_user_id', sa.Integer))


abs_table = table('absolute_modifier',
        column('id', sa.Integer),
        column('value', sa.Numeric(precision=15, scale=2)))


rel_table = table('relative_modifier',
        column('id', sa.Integer),
        column('value', sa.Numeric(precision=8, scale=5)))


def upgrade():
    op.add_column('request',
            sa.Column('payout', sa.Numeric(precision=15, scale=2), index=True,
                nullable=True))

    bind = op.get_bind()
    absolute = select([abs_table.c.value.label('value'),
                       mod_table.c.request_id.label('request_id')])\
            .select_from(join(abs_table, mod_table,
                    mod_table.c.id == abs_table.c.id))\
            .where(mod_table.c.voided_user_id == None)\
            .alias()
    relative = select([rel_table.c.value.label('value'),
                       mod_table.c.request_id.label('request_id')])\
            .select_from(join(rel_table, mod_table,
                    mod_table.c.id == rel_table.c.id))\
            .where(mod_table.c.voided_user_id == None)\
            .alias()
    abs_sum = select([request.c.id.label('request_id'),
                      request.c.base_payout.label('base_payout'),
                      func.sum(absolute.c.value).label('sum')])\
            .select_from(outerjoin(request, absolute,
                    request.c.id == absolute.c.request_id))\
            .group_by(request.c.id)\
            .alias()
    rel_sum = select([request.c.id.label('request_id'),
                      func.sum(relative.c.value).label('sum')])\
            .select_from(outerjoin(request, relative,
                    request.c.id == relative.c.request_id))\
            .group_by(request.c.id)\
            .alias()
    total_sum = select([abs_sum.c.request_id.label('request_id'),
                        ((
                            abs_sum.c.base_payout +
                            case([(abs_sum.c.sum == None, Decimal(0))],
                                    else_=abs_sum.c.sum)) *
                         (
                            1 +
                            case([(rel_sum.c.sum == None, Decimal(0))],
                                    else_=rel_sum.c.sum))).label('payout')])\
            .select_from(join(abs_sum, rel_sum,
                    abs_sum.c.request_id == rel_sum.c.request_id))
    payouts = bind.execute(total_sum)
    for request_id, payout in payouts:
        up = update(request).where(request.c.id == request_id).values(
                payout=payout)
        bind.execute(up)
    op.alter_column('request', 'payout', nullable=False,
            existing_type=sa.Numeric(precision=15, scale=2))


def downgrade():
    op.drop_column('request', 'payout')



from __future__ import absolute_import

import calendar
import datetime as dt
import re
from sqlalchemy import types
import iso8601
import iso8601.iso8601 as int_iso8601


class UTC(dt.tzinfo):

    def utcoffset(self, obj):
        return dt.timedelta()

    def dst(self, obj):
        return dt.timedelta()

    def tzname(self, obj):
        return u'UTC+00:00'


utc = UTC()


class DateTime(types.TypeDecorator):

    impl = types.DateTime

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        if dialect.name == 'mysql':
            return dt.datetime(*(value.utctimetuple()[0:6]))
        return value


def parse_datetime(date_string):
    """Parse an ISO 8601-like string into a start and end datetime.

    Parses a data and/or time string into a pair of
    :py:class:`datetime.datetime` objects representing the beginning and end
    period. A detail description of the formats accepted follows:

    The time zone is UTC, and seconds precision is not supported.
    Date ranges may be specified by two datestrings separated by an underscore.
    Dates may be written in either the basic or extended format.
    Datestrings with a leading `<` or `>` character will be parsed as a range
    starting or ending with :py:attr:`datetime.datetime.min` and
    :py:attr:`datetime.datetime.max` repsectively.

    ==============   ========================== ==========================
    String           Start Time (incl.)         End Time (incl.)
    --------------   -------------------------- --------------------------
    2015             2015-01-01T00:00           2015-01-31T23:59
    2015-02          2015-02-01T00:00           2015-02-28T23:59
    20150314         2015-03-14T00:00           2015-03-15T23:59
    20150314T15      2015-03-14T15:00           2015-03-15T15:59
    2015_2016        2015-01-01T00:00           2016-12-31T23:59
    2014-11_20150315 2014-11-01T00:00           2015-03-15T23:59
    <2015            :py:attr:`~.datetime.min`  2015-12-31T23:59
    >201408          2014-08-01T00:00           :py:attr:`~.datetime.max`
    ==============   ========================== ==========================

    :param str date_string: An ISO 8601 formatted date or date range.
    :returns: The start and end :py:class:`datetime.datetime`\s
    :rtype: tuple
    """

    def parse_start(start_string):
        return iso8601.parse_date(start_string)

    def parse_end(end_string):
        # Adapted from iso8601.parse_date, defaulting to the max instead of
        # minimum time for missing time elements.
        match = int_iso8601.ISO8601_REGEX.match(end_string)
        if match is None:
            raise iso8601.ParseError(
                    "Unable to parse date string %r" % end_string)
        groups = match.groupdict()
        year = int_iso8601.to_int(groups, 'year')
        month = int_iso8601.to_int(groups, 'month',
                default=int_iso8601.to_int(groups, 'monthdash', required=False,
                    default=dt.datetime.max.month))
        # Find the last day for the given month and year (accounting for leap
        # years).
        max_day = calendar.monthrange(year, month)[1]
        day = int_iso8601.to_int(groups, 'day',
                default=int_iso8601.to_int(groups, 'daydash', required=False,
                    default=max_day))
        hour = int_iso8601.to_int(groups, 'hour',
                default=dt.datetime.max.hour)
        minute = int_iso8601.to_int(groups, 'minute',
                default=dt.datetime.max.minute)
        second = int_iso8601.to_int(groups, 'second',
                default=dt.datetime.max.second)
        try:
            return dt.datetime(year=year, month=month, day=day, hour=hour,
                    minute=minute, second=second,
                    microsecond=dt.datetime.max.microsecond,
                    tzinfo=utc)
        # Catch ValueError, as that'd what datetime.datetime is documented to
        # raise.
        except ValueError as e:
            raise iso8601.ParseError(e)

    # Start with checking for min or max ranges
    if date_string[0] == '<':
        start_time = dt.datetime.min
        end_time = parse_end(date_string[1:])
    elif date_string[0] == '>':
        start_time = parse_start(date_string[1:])
        end_time = dt.datetime.max
    # Check for explicit ranges
    elif '_' in date_string:
        start_string, end_string = date_string.split('_', 1)
        start_time = parse_start(start_string)
        end_time = parse_end(end_string)
    # Otherwise just parse the given string as the start and end times.
    else:
        start_time = parse_start(date_string)
        end_time = parse_end(date_string)
    return start_time, end_time



#!/usr/bin/env python
from setuptools import setup, find_packages
import re
import sys
import os.path


setup_dir = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(setup_dir, 'src', 'evesrp', '__init__.py'), 'r') as f:
    init_contents = ''
    for line in f:
        init_contents += line + '\n'
version = re.search(r'^__version__ *= *u?[\'"]([^\'"]*)[\'"]', init_contents,
        re.MULTILINE)
if version:
    version = version.group(1)
else:
    raise Exception(u"Unable to find __version__ in evesrp/__init__.py")


test_requirements = [
    'beautifulsoup >= 4.4.1',
    'coverage >= 3.7.1',
    'httmock >= 1.2.2',
]
# unittest.mock was added in 3.3, but is available as a backport as the 'mock'
# package on PyPI.
if sys.version_info.major < 3 or sys.version_info.minor < 3:
    test_requirements.append('mock >= 1.0.1')


setup(
    name=u'EVE-SRP',
    version=version,
    description=u'EVE Ship Replacement Program Helper',
    author=u'Will Ross',
    author_email=u'paxswill@paxswill.com',
    url=u'https://github.com/paxswill/evesrp',
    license='BSD',
    packages=find_packages(where='src'),
    package_dir={
        '': 'src'
    },
    include_package_data=True,
    classifiers=[
        u'Development Status :: 4 - Beta',
        u'Framework :: Flask',
        u'License :: OSI Approved :: BSD License',
        u'Programming Language :: Python :: 3',
        u'Programming Language :: Python :: 2',
        u'Topic :: Games/Entertainment',
    ],
    install_requires=[
        'Babel >= 2.2.0',
        'Flask >= 0.10.1',
        'Flask-Babel >= 0.9',
        'Flask-Login >= 0.3.2',
        'Flask-Migrate ~= 1.7.0',
        'Flask-Script == 2.0.5',
        'Flask-SQLAlchemy == 2.0',
        'Flask-WTF == 0.10.2',
        'SQLAlchemy ~= 1.0.11',
        'Werkzeug >= 0.9.4',
        'WTForms >= 2.0.0',
        'alembic ~= 0.8.4',
        'requests >= 2.9.1',
        'six == 1.10.0',
        'iso8601 >= 0.1.5',
    ],
    test_suite='tests_python',
    tests_require=test_requirements,
    entry_points={
        'console_scripts': [
            'evesrp = evesrp.util.manage:main',
        ],
    },
    extras_require={
        'BraveCore': [
            'braveapi == 0.1',
            'ecdsa == 0.11',
        ],
        'OAuth': [
            'requests-oauthlib >= 0.6.0',
        ],
        'Sentry': [
            'raven[flask] >= 5.10.2',
        ],
    },
    zip_safe=False,
)



from __future__ import absolute_import
from flask import request, abort, current_app, json
import six
from sqlalchemy.orm.exc import NoResultFound

from .oauth import OAuthMethod, OAuthUser
from .. import db
from .models import Group, Pilot


class TestOAuth(OAuthMethod):

    def __init__(self, devtest=False, **kwargs):
        """:py:class:`~.AuthMethod` using TEST Auth's OAuth-based API for
        authentication and authorization.

        :param list admins: Two types of values are accepted as values in this
            list, either a string specifying a user's primary character's name,
            or their Auth ID as an integer.
        :param bool devtest: Testing parameter that changes the default domain
            for URLs from 'https://auth.pleaseignore.com' to
            'https://auth.devtest.pleaseignore.com`. Default: ``False``.
        :param str authorize_url: The URL to request OAuth authorization
            tokens. Default:
            ``'https://auth.pleaseignore.com/oauth2/authorize'``.
        :param str access_token_url: The URL for OAuth token exchange. Default:
            ``'https://auth.pleaseignore.com/oauth2/access_token'``.
        :param str base_str: The base URL for API requests. Default:
            ``'https://auth.pleaseignore.com/api/v3/'``.
        :param dict request_token_params: Additional parameters to include with
            the authorization token request. Default: ``{'scope':
            'private-read'}``.
        :param str access_token_method: HTTP Method to use for exchanging
            authorization tokens for access tokens. Default: ``'POST'``.
        :param str name: The name for this authentication method. Default:
            ``'Test OAuth'``.
        """
        if not devtest:
            domain = 'https://auth.pleaseignore.com'
        else:
            domain = 'https://auth.devtest.pleaseignore.com'
        self.base_url = domain + '/api/3.0/'
        kwargs.setdefault('authorize_url',
                domain + '/o2/authorize/')
        kwargs.setdefault('access_token_url',
                domain + '/o2/token/')
        kwargs.setdefault('refresh_token_url',
                domain + '/o2/token/')
        kwargs.setdefault('scope', ['read_profile'])
        kwargs.setdefault('method', 'POST')
        kwargs.setdefault('app_key', 'TEST_OAUTH')
        kwargs.setdefault('name', u'Test OAuth')
        super(TestOAuth, self).__init__(**kwargs)

    def _get_user_data(self):
        if not hasattr(request, '_auth_user_data'):
            resp = self.session.get(self.base_url + 'profile')
            try:
                current_app.logger.debug(u"Test OAuth API response: {}".format(
                        resp.text))
                request._auth_user_data = resp.json()
            except TypeError:
                abort(500, u"Error in receiving OAuth response: {}".format(
                        resp))
        return request._auth_user_data

    def get_user(self):
        data = self._get_user_data()
        primary_character = data[u'primary_character'][u'name']
        user_id = data[u'id']
        try:
            user = TestOAuthUser.query.filter_by(auth_id=data[u'id'],
                    authmethod=self.name).one()
            # The primary character can change
            user.name = primary_character
        except NoResultFound:
            user = TestOAuthUser(primary_character, user_id, self.name)
            db.session.add(user)
            db.session.commit()
        return user

    def is_admin(self, user):
        data = self._get_user_data()
        return super(TestOAuth, self).is_admin(user) or \
                user.auth_id in self.admins or \
                data[u'is_staff'] or \
                data[u'is_superuser']

    def get_pilots(self):
        data = self._get_user_data()
        # The Auth API will duplicate characters when there's more than one API
        # key for them.
        pilots = {}
        for character in data[u'characters']:
            pilot = Pilot.query.get(int(character[u'id']))
            if pilot is None:
                pilot = Pilot(None, character[u'name'], character[u'id'])
            pilots[character[u'id']] = pilot
        pilots = list(pilots.values())
        return pilots

    def get_groups(self):
        data = self._get_user_data()
        groups = []
        for group_info in data[u'groups']:
            group_name = group_info[u'name']
            group_id = group_info[u'id']
            try:
                group = TestOAuthGroup.query.filter_by(auth_id=group_id,
                        authmethod=self.name).one()
            except NoResultFound:
                group = TestOAuthGroup(group_name, group_id, self.name)
                db.session.add(group)
            if group.name != group_name:
                group.name = group_name
            groups.append(group)
        db.session.commit()
        return groups


class TestOAuthUser(OAuthUser):

    id = db.Column(db.Integer, db.ForeignKey(OAuthUser.id), primary_key=True)

    auth_id = db.Column(db.Integer, nullable=False, unique=True, index=True)

    def __init__(self, username, auth_id, authmethod, groups=None, **kwargs):
        self.auth_id = auth_id
        super(TestOAuthUser, self).__init__(username, authmethod, **kwargs)


class TestOAuthGroup(Group):

    id = db.Column(db.Integer, db.ForeignKey(Group.id), primary_key=True)

    auth_id = db.Column(db.Integer, nullable=False, unique=True, index=True)

    def __init__(self, name, auth_id, authmethod, **kwargs):
        self.auth_id = auth_id
        super(TestOAuthGroup, self).__init__(name, authmethod, **kwargs)



from __future__ import absolute_import
from flask import request
from sqlalchemy.orm.exc import NoResultFound

from .. import db
from .oauth import OAuthMethod, OAuthUser
from .models import Group, Pilot


class J4OAuth(OAuthMethod):

    def __init__(self, base_url='https://j4lp.com/oauth/api/v1/', **kwargs):
        """:py:class:`~.AuthMethod` for using
        `J4OAuth <https://github.com/J4LP/J4OAuth>`_ as an authentication
        source.

        :param str authorize_url: The URL to request OAuth authorization
            tokens. Default:
            ``'https://j4lp.com/oauth/authorize'``.
        :param str access_token_url: The URL for OAuth token exchange. Default:
            ``'https://j4lp.com/oauth/token'``.
        :param str base_str: The base URL for API requests. Default:
            ``'https://j4lp.com/oauth/api/v1/'``.
        :param dict request_token_params: Additional parameters to include with
            the authorization token request. Default: ``{'scope':
            ['auth_info', 'auth_groups', 'characters']}``.
        :param str access_token_method: HTTP Method to use for exchanging
            authorization tokens for access tokens. Default: ``'GET'``.
        :param str name: The name for this authentication method. Default:
            ``'J4OAuth'``.
        """
        self.base_url = base_url
        kwargs.setdefault('access_token_url', 'https://j4lp.com/oauth/token')
        kwargs.setdefault('authorize_url', 'https://j4lp.com/oauth/authorize')
        kwargs.setdefault('access_token_method', 'GET')
        kwargs.setdefault('request_token_params',
                {'scope': ['auth_info', 'auth_groups', 'characters']})
        kwargs.setdefault('name', u'J4OAuth')
        super(J4OAuth, self).__init__(**kwargs)

    def _get_user_data(self):
        if not hasattr(request, '_auth_user_data'):
            resp = self.session.get('auth_user').json()
            request._auth_user_data = resp[u'user']
        return request._auth_user_data

    def get_user(self):
        auth_user = self._get_user_data()
        try:
            user = OAuthUser.query.filter_by(name=auth_user['main_character'],
                                            authmethod=self.name).one()
        except NoResultFound:
            user = OAuthUser(name=auth_user['main_character'],
                    authmethod=self.name)
            db.session.add(user)
            db.session.commit()
        return user

    def get_pilots(self):
        pilots = []
        resp = self.session.get(self.base_url + 'characters').json()
        auth_characters = resp[u'characters']
        for character in auth_characters:
            pilot = Pilot.query.get(int(character[u'characterID']))
            if pilot is None:
                # Pass 'None' as the user, it will get set later
                pilot = Pilot(None, character[u'characterName'],
                        character[u'characterID'])
                db.session.add(pilot)
            pilots.append(pilot)
        db.session.commit()
        return pilots

    def get_groups(self):
        groups = []
        resp = self.session.get(self.base_url + 'auth_groups').json()
        auth_groups = resp[u'groups']
        # Append the user's alliance to the normal list of groups
        auth_user = self._get_user_data(token)
        auth_groups.append(u'{} alliance'.format(auth_user[u'alliance']))
        # Create/Retrieve Group objects for every group name
        for group_name in auth_groups:
            try:
                group = Group.query.filter_by(name=group_name,
                        authmethod=self.name).one()
            except NoResultFound:
                group = Group(group_name, self.name)
                db.session.add(group)
            groups.append(group)
        db.session.commit()
        return groups



#!/usr/bin.env python

from __future__ import absolute_import
from __future__ import print_function
import os
import os.path
import argparse
from itertools import cycle
import json
from decimal import Decimal
import time
import datetime as dt
import flask
import flask_script as script
from flask_migrate import Migrate, MigrateCommand, stamp
from alembic.migration import MigrationContext
from alembic.script import ScriptDirectory
import six
from .. import create_app, db, migrate, models, auth, killmail
from .datetime import utc


if six.PY3:
    unicode = str


manager = script.Manager(create_app)


class AbsolutePathAction(argparse.Action):
    """Custom argparse.Action that transforms path strings into absolute paths.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        current_absolute = os.path.abspath(os.getcwd())
        if isinstance(values, str):
            new_values = os.path.join(current_absolute, values)
        else:
            new_values = []
            for value in values:
                real_path = os.path.join(current_absolute, values)
                new_values.append(real_path)
        setattr(namespace, self.dest, new_values)


# Monkeypatch Flask-Script to consider my custom path action 'safe'
safe_actions = list(script.safe_actions)
safe_actions.append(AbsolutePathAction)
script.safe_actions = safe_actions


# Crufty workaround to get Flask-Migrate working with app factories
class MigrateManager(script.Manager):
    def __init__(self, old_command, directory='migrations'):
        super(MigrateManager, self).__init__()
        self.directory = directory
        for attr in ('app', '_commands', '_options', 'usage', 'help',
                'description', 'disable_argcomplete', 'with_default_commands',
                'parent'):
            setattr(self, attr, getattr(old_command, attr))

    def __call__(self, app=None, directory='migrations', **kwargs):
        if app is None:
            app = self.app
            if app is None:
                raise Exception(u"No app specified")

        db = kwargs.pop('db', None)

        if not isinstance(app, flask.Flask):
            app = app(**kwargs)

        # Last ditch effort to get a database handle
        if db is None:
            if 'sqlalchemy' in app.extensions:
                db = app.extensions['sqlalchemy'].db
            else:
                raise Exception(u"No database defined for app.")

        Migrate(app, db, self.directory)
        return app


migrate_path = os.path.dirname(migrate.__file__)
migrate_path = os.path.abspath(migrate_path)
migrate_manager = MigrateManager(MigrateCommand, migrate_path)
manager.add_command('db', migrate_manager)


manager.add_option('-c', '--config', dest='config', required=False,
        action=AbsolutePathAction)
manager.add_option('-i', '--instance', dest='instance_path', required=False,
        action=AbsolutePathAction)


@migrate_manager.command
def create(force=False):
    """Create tables if the database has not been configured yet."""
    # Fail if there's an alembic version set
    engine = db.get_engine(flask.current_app)
    conn = engine.connect()
    context = MigrationContext.configure(conn)
    current_rev = context.get_current_revision()
    alembic_config = flask.current_app.extensions['migrate'].migrate.get_config(
            directory=migrate_path)
    script = ScriptDirectory.from_config(alembic_config)
    latest_rev = script.get_current_head()
    if current_rev == latest_rev and not force:
        print(u"You need to run 'evesrp -c config.py db migrate' to "
              u"migrate to the latest database schema.")
    else:
        db.create_all()
        if current_rev is None:
            stamp()


@manager.shell
def shell_context():
    ctx = dict(
        app=flask.current_app,
        db=db)
    for cls in ('Request', 'Action', 'ActionType', 'Modifier',
                'AbsoluteModifier', 'RelativeModifier'):
        ctx[cls] = getattr(models, cls)
    for cls in ('Entity', 'User', 'Group', 'Note', 'APIKey', 'Permission',
            'Division', 'Pilot'):
        ctx[cls] = getattr(auth.models, cls)
    ctx['PermissionType'] = auth.PermissionType
    return ctx


class PopulatedKillmail(killmail.Killmail, killmail.RequestsSessionMixin,
        killmail.ShipNameMixin, killmail.LocationMixin):
    pass


class Populate(script.Command):
    """Populate the database with data given from a list of losses form zKB."""

    option_list = (
        script.Option('--file', '-f', dest='kill_file', required=True),
        script.Option('--users', '-u', dest='num_users', default=5, type=int),
        script.Option('--divisions', '-d', dest='num_divisions', default=3,
                type=int),
    )

    def run(self, kill_file, num_users, num_divisions, **kwargs):
        # Set up users, divisions and permissions
        users = []
        for user_num, authmethod in zip(range(num_users),
                cycle(flask.current_app.config['SRP_AUTH_METHODS'])):
            user_name = u'User {}'.format(user_num)
            user = auth.models.User.query.filter_by(name=user_name).first()
            if user is None:
                user = auth.models.User(user_name, authmethod.name)
                db.session.add(user)
            users.append(user)
        divisions = []
        for division_num in range(num_divisions):
            division_name = u'Division {}'.format(division_num)
            division = auth.models.Division.query.filter_by(
                    name=division_name).first()
            if division is None:
                division = auth.models.Division(division_name)
                db.session.add(division)
            divisions.append(division)
        db.session.add_all(divisions)
        for user in users:
            for division in divisions:
                perm = auth.models.Permission.query.filter_by(division=division,
                        permission=auth.PermissionType.submit,
                        entity=user).first()
                if perm is None:
                    auth.models.Permission(division,
                            auth.PermissionType.submit, user)
        db.session.commit()
        # load and start processing killmails
        with open(kill_file, 'r') as f:
            kills = json.load(f)
        for user, division, kill_info in zip(cycle(users), cycle(divisions), kills):
            victim = kill_info[u'victim']
            # Skip corp-level kills (towers, pocos, etc)
            if victim['characterID'] == '0':
                continue
            # make sure a Pilot exists for this killmail
            pilot = auth.models.Pilot.query.get(victim['characterID'])
            if pilot is None:
                pilot = auth.models.Pilot(user, victim['characterName'],
                        int(victim['characterID']))
            db.session.commit()
            pilot_user = pilot.user
            # create a Killmail
            args = dict(
                    kill_id=int(kill_info[u'killID']),
                    pilot_id=int(victim[u'characterID']),
                    pilot=victim[u'characterName'],
                    corp_id=int(victim[u'corporationID']),
                    corp=victim[u'corporationName'],
                    ship_id=int(victim[u'shipTypeID']),
                    system_id=int(kill_info[u'solarSystemID']),
                    verified=True)
            if victim[u'allianceID'] != '0':
                args['alliance_id'] = int(victim[u'allianceID'])
                args['alliance'] = victim[u'allianceName']
            time_struct = time.strptime(kill_info[u'killTime'], '%Y-%m-%d %H:%M:%S')
            args['timestamp'] = dt.datetime(*(time_struct[0:6]), tzinfo=utc)
            args['url'] = u'https://zkillboard.com/kill/{}'.format(
                    args['kill_id'])
            try:
                args['value'] = Decimal(kill_info[u'zkb'][u'totalValue'])
            except KeyError:
                args['value'] = Decimal(0)
            killmail = PopulatedKillmail(**args)
            try:
                killmail.ship
            except KeyError:
                continue
            # Create a request for this killmail
            models.Request(pilot_user, unicode(killmail), division, killmail)
            db.session.commit()


manager.add_command('populate', Populate())


def main():
    manager.run()


if __name__ == '__main__':
    main()



"""Name constraints and indicies in a common manner across databases.

Revision ID: 108d7a5881c7
Revises: 3d35f833a24
Create Date: 2014-07-22 13:19:50.598892

"""

# revision identifiers, used by Alembic.
revision = '108d7a5881c7'
down_revision = '3d35f833a24'

from alembic import op
import sqlalchemy as sa
import six
from itertools import groupby


fkeys = {
    ('absolute_modifier', 'id'): ('modifier', 'id'),
    ('action', 'request_id'): ('request', 'id'),
    ('action', 'user_id'): ('user', 'id'),
    ('apikey', 'user_id'): ('user', 'id'),
    ('core_group', 'id'): ('group', 'id'),
    ('core_user', 'id'): ('user', 'id'),
    ('group', 'id'): ('entity', 'id'),
    ('modifier', 'request_id'): ('request', 'id'),
    ('modifier', 'user_id'): ('user', 'id'),
    ('modifier', 'voided_user_id'): ('user', 'id'),
    ('note', 'user_id'): ('user', 'id'),
    ('note', 'noter_id'): ('user', 'id'),
    ('permission', 'division_id'): ('division', 'id'),
    ('permission', 'entity_id'): ('entity', 'id'),
    ('pilot', 'user_id'): ('user', 'id'),
    ('relative_modifier', 'id'): ('modifier', 'id'),
    ('request', 'submitter_id'): ('user', 'id'),
    ('request', 'division_id'): ('division', 'id'),
    ('request', 'pilot_id'): ('pilot', 'id'),
    ('test_group', 'id'): ('group', 'id'),
    ('test_user', 'id'): ('user', 'id'),
    ('transformerref', 'division_id'): ('division', 'id'),
    ('user', 'id'): ('entity', 'id'),
    ('users_groups', 'user_id'): ('user', 'id'),
    ('users_groups', 'group_id'): ('group', 'id'),
}

pkeys = {
    'absolute_modifier': ['id'],
    'action': ['id'],
    'apikey': ['id'],
    'core_group': ['id'],
    'core_user': ['id'],
    'division': ['id'],
    'entity': ['id'],
    'group': ['id'],
    'modifier': ['id'],
    'note': ['id'],
    'permission': ['id'],
    'pilot': ['id'],
    'relative_modifier': ['id'],
    'request': ['id'],
    'test_group': ['id'],
    'test_user': ['id'],
    'transformerref': ['id'],
    'user': ['id'],
}

uniq = {
    'permission': ('division_id', 'entity_id', 'permission'),
    'transformerref': ('division_id', 'attribute_name'),
}


naming_schemes = {
    'standard': {
        'foreignkey': ('fk_{local_table}_{local_col}_{remote_table}'
                       '_{remote_col}'),
        'primary': 'pk_{table_name}',
        'unique': 'uq_{table_name}_{columns[0]}',
    },
    'mysql': {
        'foreignkey': '{local_table}_ibfk_{table_num}',
        'primary': 'PRIMARY',
        'unique': '{columns[0]}',
    },
    'postgresql': {
        'foreignkey': '{local_table}_{local_col}_fkey',
        'primary': '{table_name}_pkey',
        'unique': None,
    },
}


def prune_tablenames():
    # Skip tables that aren't present in the target database, like specific
    # AuthMethods that aren't enabled.
    metadata = sa.MetaData(bind=op.get_bind())
    metadata.reflect()
    tables = {t.name for t in metadata.sorted_tables}
    for table_col in list(six.iterkeys(fkeys)):
        if table_col[0] not in tables:
            del fkeys[table_col]
    for table in list(six.iterkeys(pkeys)):
        if table not in tables:
            del pkeys[table]
    for table in list(six.iterkeys(uniq)):
        if table not in tables:
            del uniq[table]


def upgrade():
    prune_tablenames()
    bind = op.get_bind()
    # Add new constraints with new names and remove old constrains with the
    # old, DB-specific names
    # foreign keys
    for local, remote in six.iteritems(fkeys):
        # Give None as the name to use the predefined naming scheme
        op.create_foreign_key(None, local[0], remote[0], [local[1]],
                [remote[1]])
    if bind.dialect.name == 'mysql':
        # foreign keys
        grouped_table_names = groupby(sorted(fkeys.keys()), lambda k: k[0])
        counted_table_names = [(i, len(list(j))) for i, j in
                grouped_table_names]
        for table_name, count in counted_table_names:
            for c in range(1, count+1):
                constraint_name = naming_schemes['mysql']['foreignkey'].format(
                        local_table=table_name,
                        table_num=c)
                op.drop_constraint(constraint_name, table_name,
                        type_='foreignkey')
    elif bind.dialect.name == 'postgresql':
        # foreign keys
        for table, column in fkeys.keys():
            constraint_name = naming_schemes['postgresql']['foreignkey'].\
                    format(local_table=table, local_col=column)
            op.drop_constraint(constraint_name, table)
    # primary keys
    # MySQL names all primary keys 'PRIMARY', and doesn't let you change them.
    if bind.dialect.name != 'mysql':
        for table, columns in pkeys:
            op.create_primary_key(None, table, columns)
            old_name = naming_schemes[bind.dialect.name]['primary'].format(
                    table_name=table)
            op.drop_constraint(old_name, table, type_='primary')
    # unique
    for table, columns in six.iteritems(uniq):
        op.create_unique_constraint(None, table, columns)
        if bind.dialect.name == 'postgresql':
            old_name = '{}_{}_key'.format(table, '_'.join(columns))
        else:
            old_name = naming_schemes[bind.dialect.name]['unique'].format(
                    table_name=table, columns=columns)
        op.drop_constraint(old_name, table, type_='unique')


def downgrade():
    prune_tablenames()
    # Add all the DB specific names back and remove the standard names
    bind = op.get_bind()
    # foreign keys
    counts = {}
    for local, remote in six.iteritems(fkeys):
        # add old
        current_count = counts.get(local[0], 1)
        old_name = naming_schemes[bind.dialect.name]['foreignkey'].format(
                local_table=local[0], local_col=local[1],
                remote_table=remote[0], remote_col=remote[1],
                table_num=current_count)
        op.create_foreign_key(old_name, local[0], remote[0],
                [local[1]], [remote[1]])
        # remove new
        new_name = naming_schemes['standard']['foreignkey'].format(
                local_table=local[0], local_col=local[1],
                remote_table=remote[0], remote_col=remote[1],
                table_num=current_count)
        op.drop_constraint(new_name, local[0], type_='foreignkey')
        counts[local[0]] = current_count + 1
    # primary keys
    if bind.dialect.name != 'mysql':
        for table, columns in pkeys:
            old_name = naming_schemes[bind.dialect.name]['primary'].format(
                    table_name=table)
            new_name = naming_schemes['standard']['primary'].format(
                    table_name=table)
            op.create_primary_key(old_name, table, columns)
            op.drop_constraint(new_name, table, type_='primary')

    # unique
    for table, columns in six.iteritems(uniq):
        if bind.dialect.name == 'postgresql':
            old_name = '{}_{}_key'.format(table, '_'.join(columns))
        else:
            old_name = naming_schemes[bind.dialect.name]['unique'].format(
                    table_name=table, columns=columns)
        op.create_unique_constraint(old_name, table, columns)
        new_name = naming_schemes['standard']['unique'].format(
                table_name=table, columns=columns)
        op.drop_constraint(new_name, table, type_='unique')



"""Change to a division-specified transformer for ships and pilots.

Revision ID: c1fc69b629
Revises: 2f22504b1e6
Create Date: 2014-05-07 15:10:10.404234

"""

# revision identifiers, used by Alembic.
revision = 'c1fc69b629'
down_revision = '2f22504b1e6'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('division',
            sa.Column('pilot_transformer', sa.PickleType(), nullable=True))
    op.add_column('division',
            sa.Column('ship_transformer', sa.PickleType(), nullable=True))
    op.drop_column('request', 'ship_url')


def downgrade():
    op.add_column('request',
            sa.Column('ship_url', sa.VARCHAR(length=500), nullable=True))
    op.drop_column('division', 'ship_transformer')
    op.drop_column('division', 'pilot_transformer')



"""Rename action._type to action.type_

Revision ID: 45024170cf6
Revises: 337978f8c75
Create Date: 2014-06-18 14:21:37.202030

"""

# revision identifiers, used by Alembic.
revision = '45024170cf6'
down_revision = '337978f8c75'

from alembic import op
import sqlalchemy as sa
from evesrp.models import ActionType


def upgrade():
    op.alter_column('action',
            column_name='_type',
            new_column_name='type_',
            existing_type=ActionType.db_type,
            existing_nullable=False)


def downgrade():
    op.alter_column('action',
            column_name='type_',
            new_column_name='_type',
            existing_type=ActionType.db_type,
            existing_nullable=False)



"""Discern between access and refresh tokens for OAuth

Revision ID: 3f453bb4f2d7
Revises: e6b2c156b11
Create Date: 2016-02-04 15:58:56.096867

"""

# revision identifiers, used by Alembic.
revision = '3f453bb4f2d7'
down_revision = 'e6b2c156b11'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # Inspect the current database to see if we're using OAuth
    metadata = sa.MetaData(bind=op.get_bind())
    metadata.reflect()
    table_names = {t.name for t in metadata.sorted_tables}
    if 'o_auth_user' in table_names:
        op.add_column('o_auth_user',
                sa.Column('access_expiration',
                    sa.DateTime(),
                    nullable=True))
        op.add_column('o_auth_user',
                sa.Column('refresh_token',
                    sa.String(length=100, convert_unicode=True),
                    nullable=True))
        op.alter_column('o_auth_user', 'token', new_column_name='access_token',
                existing_type=sa.String(length=100, convert_unicode=True),
                existing_nullable=True)


def downgrade():
    # Inspect the current database to see if we're using OAuth
    metadata = sa.MetaData(bind=op.get_bind())
    metadata.reflect()
    table_names = {t.name for t in metadata.sorted_tables}
    if 'o_auth_user' in table_names:
        op.drop_column('o_auth_user', 'refresh_token')
        op.drop_column('o_auth_user', 'access_expiration')
        op.alter_column('o_auth_user', 'access_token', new_column_name='token',
                existing_type=sa.String(length=100, convert_unicode=True),
                existing_nullable=True)



from __future__ import absolute_import
from flask import request
from flask.json import JSONEncoder
import six
from decimal import Decimal
from .util import PrettyDecimal

if six.PY3:
    unicode = str


class IterableEncoder(JSONEncoder):
    def default(self, o):
        try:
            iterator = iter(o)
        except TypeError:
            pass
        else:
            return list(o)
        return super(IterableEncoder, self).default(o)


class PrivateJsonEncoder(JSONEncoder):
    def default(self, o):
        if hasattr(o, '_json'):
            try:
                extended = request.json_extended
            except AttributeError:
                extended = request.args.get('extended', False)
            # request.json_extended can be a dict to set extended mode for only
            # certain types of objects
            if isinstance(extended, dict):
                scope_class = object
                scope_extended = False
                for cls, ext in six.iteritems(extended):
                    if isinstance(o, cls) and issubclass(type(o), scope_class):
                        scope_extended = ext
                        scope_class = cls
                extended = scope_extended
            return o._json(extended)
        return super(PrivateJsonEncoder, self).default(o)


class DecimalEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, PrettyDecimal):
            return o.currency()
        elif isinstance(o, Decimal):
            return unicode(o)
        return super(DecimalEncoder, self).default(o)


# Multiple inheritance FTW
class SRPEncoder(PrivateJsonEncoder, IterableEncoder, DecimalEncoder):
    pass



from __future__ import absolute_import
from decimal import Decimal
import six
from sqlalchemy.types import TypeDecorator, Numeric


if six.PY3:
    unicode = str


class PrettyDecimal(Decimal):
    """:py:class:`~.Decimal` subclass that can pretty-print itself."""

    def currency(self):
        """Format the Decimal as a currency number.

        An older implementation of this also handled formatting with thousands
        separators. This is not longer supported.
        """
        sign, digits, exp = self.quantize(Decimal('0.01')).as_tuple()
        digits = list(map(unicode, digits))
        result = []
        for i in range(2):
            result.append(digits.pop() if digits else u'0')
        result.append(u'.')
        if not digits:
            result.append(u'0')
        while digits:
            result.append(digits.pop())
        result.append(u'-' if sign else u'')
        return u''.join(reversed(result))


class PrettyNumeric(TypeDecorator):
    """Type Decorator for :py:class:`~.Numeric` that reformats the values into
    :py:class:`PrettyDecimal`\s.
    """

    impl = Numeric

    def process_result_value(self, value, dialect):
        return PrettyDecimal(value) if value is not None else None



import six
from babel import numbers, Locale
from flask import current_app
from flask_babel import format_datetime, get_locale
from .. import babel


if six.PY3:
    unicode = str


def currencyfmt(currency):
    # We're only formatting ISK, so it has a bit of a special format
    number = numbers.format_decimal(currency, format="#,##0.00;-#",
            locale=get_locale())
    return number


def numberfmt(number):
    return numbers.format_number(number, locale=get_locale())


def percentfmt(percent):
    return numbers.format_percent(percent, locale=get_locale())


def enabled_locales():
    all_locales = [(unicode(l), l) for l in babel.list_translations()]
    enabled_locales = current_app.config.get('SRP_LOCALES', [])
    if not enabled_locales:
        for locale in all_locales:
            yield locale[0]
    else:
        for enabled_locale in enabled_locales:
            for locale_str, locale in all_locales:
                if locale_str == enabled_locale:
                    yield locale
